import itertools
import math
import re
from functools import reduce
from typing import Callable, TypeVar, Optional, Iterable, Tuple, Set, List, Dict, Sequence

import isla.language
from pathos import multiprocessing as pmp

S = TypeVar("S")
T = TypeVar("T")


def is_int(maybe_int: str) -> bool:
    try:
        int(maybe_int)
        return True
    except ValueError:
        return False


def is_float(maybe_float: str) -> bool:
    try:
        f = float(maybe_float)
        return not math.isinf(f) and not math.isnan(f)
    except ValueError:
        return False


def e_assert_present(expression: T, message: Optional[str] = None) -> T:
    return e_assert(expression, lambda e: e is not None, message)


def e_assert(expression: T, assertion: Callable[[T], bool], message: Optional[str] = None) -> T:
    assert assertion(expression), message or ""
    return expression


def parallel_all(f: Callable[[T], bool], iterable: Iterable[T], chunk_size=16, processes=pmp.cpu_count()) -> bool:
    # l = list(iterable)
    # chunked_list = [l[i:i + chunk_size] for i in range(0, len(l), chunk_size)]
    #
    # idx = 0
    # while idx < len(chunked_list):
    #     with pmp.ProcessingPool(processes=processes) as pool:
    #         if not all(pool.map(lambda chunk: all(f(elem) for elem in chunk), chunked_list[idx:idx + processes])):
    #             return False
    #
    #     idx += processes
    #
    # return True
    with pmp.ProcessingPool(processes=processes) as pool:
        results = pool.imap(f, iterable)
        for result in results:
            if not result:
                return False

    return True


def parallel_any(f: Callable[[T], bool], iterable: Iterable[T], processes=pmp.cpu_count()) -> bool:
    with pmp.ProcessingPool(processes=processes) as pool:
        results = pool.uimap(f, iterable)
        for result in results:
            if result:
                return True

    return False


def mappings(a: Iterable[S], b: Iterable[T]) -> List[Dict[S, T]]:
    a = list(a)
    return [
        dict(zip(a, item))
        for item in itertools.product(b, repeat=len(a))]


def connected_chains(relation: Iterable[Tuple[S, S]]) -> Set[Tuple[S, ...]]:
    result: Set[Tuple[S, ...]] = set(relation)
    while True:
        matches: Set[Tuple[Tuple[S, ...], Tuple[S, ...]]] = {
            (chain_1, chain_2)
            for chain_1 in result
            for chain_2 in result
            if chain_1[-1] == chain_2[0]
        }

        new_chains: Set[Tuple[S, ...]] = {
            chain_1 + chain_2[1:]
            for chain_1, chain_2 in matches
        }

        result_until_now = result | new_chains

        if result_until_now == result:
            break

        result = result_until_now

    # Exclude subsequences. Cannot do this before, since we need the original
    # chains for creating longer ones.
    return {
        chain for idx, chain in enumerate(result)
        if not any(
            idx != other_idx and
            len(other_chain) > len(chain) and
            any(
                other_chain[idx_1:idx_2] == chain
                for idx_1 in range(len(other_chain))
                for idx_2 in range(idx_1, len(other_chain) + 1)
            )
            for other_idx, other_chain in enumerate(result)
        )
    }


def transitive_closure(relation: Iterable[Tuple[S, T]]) -> Set[Tuple[S, T]]:
    closure = set(relation)
    while True:
        new_relations: Set[Tuple[S, T]] = {
            (x, w)
            for x, y in closure
            for q, w in closure if q == y}

        closure_until_now = closure | new_relations

        if closure_until_now == closure:
            break

        closure = closure_until_now

    return closure


def non_consecutive_ordered_sub_sequences(sequence: Iterable[T], length: int) -> Set[Tuple[T, ...]]:
    sequence = list(sequence)
    if len(sequence) < length:
        return set()

    filters_list = [f for f in list(itertools.product([0, 1], repeat=len(sequence))) if sum(f) == length]

    return {tuple(itertools.compress(sequence, a_filter)) for a_filter in filters_list}


def all_interleavings(a: Sequence[S], b: Sequence[T]) -> List[List[S | T]]:
    slots = [None] * (len(a) + len(b))
    for splice in itertools.combinations(range(0, len(slots)), len(b)):
        it_b = iter(b)
        for s in splice:
            slots[s] = next(it_b)
        it_a = iter(a)
        slots = [e if e else next(it_a) for e in slots]
        yield slots
        slots = [None] * (len(slots))

    return slots


def construct_multiple(constructor: Callable[..., T], args: List[Iterable]) -> List[T]:
    """
    :param constructor: Constructor to apply to individual arguments.
    :param args: List of different argument options. Length of outer list must match cardinality of constructor.
    :return: Instantiated objects for the different possibilities.
    """

    return [constructor(*combination) for combination in itertools.product(*args)]


def reduce_multiple(function: Callable[[T], T], sequence: List[Iterable[T]]) -> List[T]:
    return [reduce(function, combination) for combination in itertools.product(*sequence)]


def replace_formula_by_formulas(
        in_formula: isla.language.Formula,
        to_replace: Callable[[isla.language.Formula], Optional[Iterable[isla.language.Formula]]]
) -> List[isla.language.Formula]:
    """
    Replaces a formula inside a conjunction or disjunction.
    to_replace is either (1) a formula to replace, or (2) a predicate which holds if the given formula
    should been replaced (if it returns True, replace_with must not be None), or (3) a function returning
    the formula to replace if the subformula should be replaced, or False otherwise. For (3), replace_with
    may be None (it is irrelevant).
    """

    maybe_results = to_replace(in_formula)
    if maybe_results is not None:
        return [r for result in maybe_results for r in replace_formula_by_formulas(result, to_replace)]

    if isinstance(in_formula, isla.language.ConjunctiveFormula):
        return reduce_multiple(lambda a, b: a & b, [
            replace_formula_by_formulas(child, to_replace)
            for child in in_formula.args])
    elif isinstance(in_formula, isla.language.DisjunctiveFormula):
        return reduce_multiple(lambda a, b: a | b, [
            replace_formula_by_formulas(child, to_replace)
            for child in in_formula.args])
    elif isinstance(in_formula, isla.language.NegatedFormula):
        return [
            -child_result
            for child_result in replace_formula_by_formulas(in_formula.args[0], to_replace)
        ]
    elif isinstance(in_formula, isla.language.ForallFormula):
        in_formula: isla.language.ForallFormula
        return construct_multiple(
            lambda inner_formula: isla.language.ForallFormula(
                in_formula.bound_variable,
                in_formula.in_variable,
                inner_formula,
                in_formula.bind_expression,
                in_formula.already_matched,
                id=in_formula.id
            ), [replace_formula_by_formulas(in_formula.inner_formula, to_replace)])
    elif isinstance(in_formula, isla.language.ExistsFormula):
        in_formula: isla.language.ExistsFormula
        return construct_multiple(
            lambda inner_formula: isla.language.ExistsFormula(
                in_formula.bound_variable,
                in_formula.in_variable,
                inner_formula,
                in_formula.bind_expression),
            [replace_formula_by_formulas(in_formula.inner_formula, to_replace)])
    elif isinstance(in_formula, isla.language.ExistsIntFormula):
        in_formula: isla.language.ExistsIntFormula
        return construct_multiple(
            lambda inner_formula:
            isla.language.ExistsIntFormula(
                in_formula.bound_variable,
                inner_formula),
            [replace_formula_by_formulas(in_formula.inner_formula, to_replace)])
    elif isinstance(in_formula, isla.language.ForallIntFormula):
        in_formula: isla.language.ForallIntFormula
        return construct_multiple(
            lambda inner_formula:
            isla.language.ForallIntFormula(
                in_formula.bound_variable,
                inner_formula),
            [replace_formula_by_formulas(in_formula.inner_formula, to_replace)])

    return [in_formula]


def tree_in(tree: isla.language.DerivationTree, iterable: Iterable[isla.language.DerivationTree]) -> bool:
    return any(tree.structurally_equal(t) for t in iterable)


def remove_spaces(inp: str) -> str:
    return re.sub(r"\s+", "", str(inp))


def try_except(success: Callable[[], T], failure: Callable[[], T] | T, *exceptions) -> S | T:
    # See https://stackoverflow.com/questions/7088009/python-try-except-as-an-expression
    try:
        return success()
    except exceptions or Exception:
        return failure() if callable(failure) else failure



import re
from abc import ABC
from dataclasses import dataclass
from typing import Set, Optional, Callable, List, Tuple, cast, Dict

import antlr4
import z3
from antlr4 import InputStream
from isla import language
from isla.isla_predicates import STANDARD_STRUCTURAL_PREDICATES, STANDARD_SEMANTIC_PREDICATES
from isla.language import ISLaEmitter, StructuralPredicate, SemanticPredicate, VariableManager, Variable, Formula, \
    parse_tree_text, antlr_get_text_with_whitespace, ISLaUnparser, MExprEmitter, BailPrintErrorStrategy, \
    ensure_unique_bound_variables, VariablesCollector
from isla.type_defs import Grammar
from isla.z3_helpers import get_symbols, smt_expr_to_str
from orderedset import OrderedSet

from islearn.isla_language import IslaLanguageListener
from islearn.islearn_predicates import INTERNET_CHECKSUM_PREDICATE, HEX_TO_DEC_PREDICATE
from islearn.isla_language.IslaLanguageLexer import IslaLanguageLexer
from islearn.isla_language.IslaLanguageParser import IslaLanguageParser
from islearn.mexpr_lexer.MexprLexer import MexprLexer
from islearn.mexpr_parser import MexprParserListener
from islearn.mexpr_parser.MexprParser import MexprParser

NONTERMINAL_PLACEHOLDER = "<?NONTERMINAL>"
MEXPR_PLACEHOLDER = "<?MATCHEXPR>"
STRING_PLACEHOLDER = "<?STRING>"
DSTRINGS_PLACEHOLDER = "<?DSTRINGS>"


class PlaceholderVariable(language.BoundVariable, ABC):
    pass


@dataclass(frozen=True, eq=True, init=True)
class NonterminalPlaceholderVariable(PlaceholderVariable):
    name: str
    n_type: str = NONTERMINAL_PLACEHOLDER


@dataclass(frozen=True, eq=True, init=True)
class NonterminalStringPlaceholderVariable(PlaceholderVariable):
    name: str

    def __str__(self):
        return NONTERMINAL_PLACEHOLDER


@dataclass(frozen=True, eq=True, init=True)
class StringPlaceholderVariable(PlaceholderVariable):
    name: str

    def __str__(self):
        return STRING_PLACEHOLDER


@dataclass(frozen=True, eq=True, init=True)
class DisjunctiveStringsPlaceholderVariable(PlaceholderVariable):
    name: str

    def __str__(self):
        return DSTRINGS_PLACEHOLDER


StringPlaceholderVariableTypes = StringPlaceholderVariable | DisjunctiveStringsPlaceholderVariable


@dataclass(frozen=True, eq=True, init=True)
class MexprPlaceholderVariable(PlaceholderVariable):
    name: str
    variables: Tuple[Variable]
    n_type: str = MEXPR_PLACEHOLDER

    def substitute_variables(self, subst_map: Dict[Variable, Variable]) -> 'MexprPlaceholderVariable':
        return MexprPlaceholderVariable(
            self.name,
            tuple([cast(Variable, subst_map.get(var, var)) for var in self.variables])
        )

    def __str__(self):
        def arg_to_str(v: Variable):
            if isinstance(v, NonterminalPlaceholderVariable):
                return v.name
            assert not isinstance(v, PlaceholderVariable)
            return f"{v.n_type} {v.name}"

        return MEXPR_PLACEHOLDER[:-1] + "(" + ", ".join(map(arg_to_str, self.variables)) + ")>"


class AbstractBindExpression(language.BindExpression):
    def __init__(self, *bound_elements: MexprPlaceholderVariable | str | language.BoundVariable | List[str]):
        super().__init__(*bound_elements)
        assert sum(bool(isinstance(elem, MexprPlaceholderVariable)) for elem in bound_elements) <= 1

        # Below assignment is for type checking reasons only
        self.bound_elements: List[
            MexprPlaceholderVariable |
            language.BoundVariable |
            List[language.BoundVariable]] = self.bound_elements

    def bound_variables(self) -> OrderedSet[language.BoundVariable]:
        # Not isinstance(var, BoundVariable) since we want to exclude dummy variables
        result = OrderedSet([var for var in self.bound_elements if type(var) is language.BoundVariable])
        result.update([
            var for mexpr_ph in self.bound_elements
            if isinstance(mexpr_ph, MexprPlaceholderVariable)
            for var in mexpr_ph.variables
            if isinstance(var, language.BoundVariable)])
        return result

    def substitute_variables(self, subst_map: Dict[Variable, Variable]) -> 'AbstractBindExpression':
        return AbstractBindExpression(
            *[elem if isinstance(elem, list)
              else (
                elem.substitute_variables(subst_map)
                if isinstance(elem, MexprPlaceholderVariable)
                else subst_map.get(elem, elem))
              for elem in self.bound_elements])

    def __str__(self):
        return ''.join(map(
            lambda e: f'{str(e)}'
            if isinstance(e, str)
            else ("[" + "".join(map(str, e)) + "]") if isinstance(e, list)
            else (f"{{{'' if isinstance(e, MexprPlaceholderVariable) else e.n_type + ' '}{str(e)}}}"
                  if not isinstance(e, language.DummyVariable)
                  else (str(e))), self.bound_elements))


class AbstractVariableManager(VariableManager):
    def __init__(self, grammar: Optional[Grammar] = None):
        super().__init__(grammar)

    def _var(self,
             name: str,
             n_type: Optional[str],
             constr: Optional[Callable[[str, Optional[str]], Variable]] = None) -> Variable:
        if n_type is not None:
            assert (self.grammar is None or
                    n_type == Variable.NUMERIC_NTYPE or
                    n_type in self.grammar or
                    n_type == NONTERMINAL_PLACEHOLDER or
                    n_type == STRING_PLACEHOLDER or
                    n_type == DSTRINGS_PLACEHOLDER), \
                f"Unknown nonterminal type {n_type} for variable {name}"

        try:
            return next(var for var_name, var in self.variables.items() if var_name == name)
        except StopIteration:
            pass

        if n_type == NONTERMINAL_PLACEHOLDER:
            return self.variables.setdefault(name, NonterminalPlaceholderVariable(name))

        if n_type == STRING_PLACEHOLDER:
            return self.variables.setdefault(name, StringPlaceholderVariable(name))

        if n_type == DSTRINGS_PLACEHOLDER:
            return self.variables.setdefault(name, DisjunctiveStringsPlaceholderVariable(name))

        if constr is not None and n_type:
            return self.variables.setdefault(name, constr(name, n_type))

        matching_placeholders = [var for var_name, var in self.placeholders.items() if var_name == name]
        if matching_placeholders:
            return matching_placeholders[0]

        assert constr is not None
        return self.placeholders.setdefault(name, constr(name, None))

    def fresh_name(self, prefix: str = "v") -> str:
        idx = 0
        while f"{prefix}_{idx}" in self.placeholders or f"{prefix}_{idx}" in self.variables:
            idx += 1

        return f"{prefix}_{idx}"


def used_variables_in_concrete_syntax(inp: str | IslaLanguageParser.StartContext) -> OrderedSet[str]:
    if isinstance(inp, str):
        lexer = IslaLanguageLexer(InputStream(inp))
        parser = IslaLanguageParser(antlr4.CommonTokenStream(lexer))
        parser._errHandler = BailPrintErrorStrategy()
        context = parser.start()
    else:
        assert isinstance(inp, IslaLanguageParser.StartContext)
        context = inp

    collector = ConcreteSyntaxUsedVariablesCollector()
    antlr4.ParseTreeWalker().walk(collector, context)
    return collector.used_variables


class ConcreteSyntaxUsedVariablesCollector(IslaLanguageListener.IslaLanguageListener):
    def __init__(self):
        self.used_variables: OrderedSet[str] = OrderedSet()

    def collect_used_variables_in_mexpr(self, inp: str) -> None:
        lexer = MexprLexer(InputStream(inp))
        parser = MexprParser(antlr4.CommonTokenStream(lexer))
        parser._errHandler = BailPrintErrorStrategy()
        collector = ConcreteSyntaxMexprUsedVariablesCollector()
        antlr4.ParseTreeWalker().walk(collector, parser.matchExpr())
        self.used_variables.update(collector.used_variables)

    def enterForall(self, ctx: IslaLanguageParser.ForallContext):
        if ctx.varId:
            self.used_variables.add(parse_tree_text(ctx.varId))

    def enterForallMexpr(self, ctx: IslaLanguageParser.ForallMexprContext):
        if ctx.varId:
            self.used_variables.add(parse_tree_text(ctx.varId))
        self.collect_used_variables_in_mexpr(antlr_get_text_with_whitespace(ctx.STRING())[1:-1])

    def enterExists(self, ctx: IslaLanguageParser.ExistsContext):
        if ctx.varId:
            self.used_variables.add(parse_tree_text(ctx.varId))

    def enterExistsMexpr(self, ctx: IslaLanguageParser.ExistsMexprContext):
        if ctx.varId:
            self.used_variables.add(parse_tree_text(ctx.varId))
        self.collect_used_variables_in_mexpr(antlr_get_text_with_whitespace(ctx.STRING())[1:-1])


class ConcreteSyntaxMexprUsedVariablesCollector(MexprParserListener.MexprParserListener):
    def __init__(self):
        self.used_variables: OrderedSet[str] = OrderedSet()

    def enterMatchExprVar(self, ctx: MexprParser.MatchExprVarContext):
        self.used_variables.add(parse_tree_text(ctx.ID()))


class AbstractISLaEmitter(ISLaEmitter):
    def __init__(
            self,
            grammar: Optional[Grammar] = None,
            structural_predicates: Set[StructuralPredicate] = STANDARD_STRUCTURAL_PREDICATES,
            semantic_predicates: Set[SemanticPredicate] = STANDARD_SEMANTIC_PREDICATES):
        super().__init__(grammar, structural_predicates, semantic_predicates)

        self.mgr = AbstractVariableManager(grammar)
        self.next_nonterminal_string_placeholder_index = 1
        self.next_string_placeholder_index = 1
        self.next_dstrings_placeholder_index = 1

    def enterStart(self, ctx: IslaLanguageParser.StartContext):
        self.used_variables = used_variables_in_concrete_syntax(ctx)

    def exitStart(self, ctx: IslaLanguageParser.StartContext):
        try:
            formula: Formula = self.mgr.create(self.formulas[ctx.formula()])
            formula = ensure_unique_bound_variables(formula)
            self.used_variables.update({var.name for var in VariablesCollector.collect(formula)})
            self.result = \
                self.close_over_xpath_expressions(
                    self.close_over_free_nonterminals(formula))

            # NOTE: We're skipping the free variables check from the standard `ISLaEmitter`
            #       since placeholders and variables in match expression placeholders would
            #       also show up (the latter would be a little tricky to distinguish from
            free_variables = [
                var for var in self.result.free_variables()
                if not isinstance(var, language.Constant)
                   and not isinstance(var, PlaceholderVariable)]
            if free_variables:
                raise SyntaxError('Unbound variables: ' + ', '.join(map(str, free_variables)))
        except RuntimeError as exc:
            raise SyntaxError(str(exc))

    def exitPredicateArg(self, ctx: IslaLanguageParser.PredicateArgContext):
        text = parse_tree_text(ctx)

        if ctx.ID():
            self.predicate_args[ctx] = self.get_var(text)
        elif ctx.INT():
            self.predicate_args[ctx] = int(text)
        elif ctx.STRING():
            self.predicate_args[ctx] = text[1:-1]
        elif ctx.VAR_TYPE():
            variable = self.register_var_for_free_nonterminal(parse_tree_text(ctx.VAR_TYPE()))
            self.predicate_args[ctx] = variable
        elif ctx.XPATHEXPR():
            variable = self.register_var_for_xpath_expression(parse_tree_text(ctx))
            self.predicate_args[ctx] = variable
        elif text == STRING_PLACEHOLDER:
            self.predicate_args[ctx] = StringPlaceholderVariable(
                f"STRING_{self.next_string_placeholder_index}")
            self.next_string_placeholder_index += 1
        elif text == DSTRINGS_PLACEHOLDER:
            self.predicate_args[ctx] = StringPlaceholderVariable(
                f"DSTRINGS_{self.next_dstrings_placeholder_index}")
            self.next_dstrings_placeholder_index += 1
        elif text == NONTERMINAL_PLACEHOLDER:
            self.predicate_args[ctx] = NonterminalStringPlaceholderVariable(
                f"NONTERMINAL_{self.next_nonterminal_string_placeholder_index}")
            self.next_nonterminal_string_placeholder_index += 1
        else:
            assert False, f"Unknown predicate argument type: {text}"

    def exitSMTFormula(self, ctx: IslaLanguageParser.SMTFormulaContext):
        formula_text = self.smt_expressions[ctx.sexpr()]

        match = re.search("(" + re.escape(STRING_PLACEHOLDER) + ")", formula_text)
        if match:
            for group_idx in range(1, 1 + len(match.groups())):
                new_var = self.mgr.bv(self.mgr.fresh_name("STRING"), STRING_PLACEHOLDER)
                fr, to = match.span(group_idx)
                formula_text = formula_text[:fr] + new_var.name + formula_text[to:]

        match = re.search("(" + re.escape(DSTRINGS_PLACEHOLDER) + ")", formula_text)
        if match:
            for group_idx in range(1, 1 + len(match.groups())):
                new_var = self.mgr.bv(self.mgr.fresh_name("DSTRINGS"), DSTRINGS_PLACEHOLDER)
                fr, to = match.span(group_idx)
                formula_text = formula_text[:fr] + new_var.name + formula_text[to:]

        try:
            z3_constr = z3.parse_smt2_string(
                f"(assert {formula_text})",
                decls=({var: z3.String(var)
                        for var in self.known_var_names()} |
                       {nonterminal: z3.String(var.name)
                        for nonterminal, var in self.vars_for_free_nonterminals.items()} |
                       {xpath_expr: z3.String(var.name)
                        for xpath_expr, var in self.vars_for_xpath_expressions.items()}))[0]
        except z3.Z3Exception as exp:
            raise SyntaxError(
                f"Error parsing SMT formula '{formula_text}', {exp.value.decode().strip()}")

        free_vars = [self.get_var(str(s)) for s in get_symbols(z3_constr)]
        self.formulas[ctx] = language.SMTFormula(z3_constr, *free_vars)

    def parse_mexpr(self, inp: str, mgr: VariableManager) -> language.BindExpression:
        class BailPrintErrorStrategy(antlr4.BailErrorStrategy):
            def recover(self, recognizer: antlr4.Parser, e: antlr4.RecognitionException):
                recognizer._errHandler.reportError(recognizer, e)
                super().recover(recognizer, e)

        lexer = MexprLexer(InputStream(inp))
        parser = MexprParser(antlr4.CommonTokenStream(lexer))
        parser._errHandler = BailPrintErrorStrategy()
        mexpr_emitter = AbstractMExprEmitter(mgr)
        antlr4.ParseTreeWalker().walk(mexpr_emitter, parser.matchExpr())
        return AbstractBindExpression(*mexpr_emitter.result)


class AbstractMExprEmitter(MExprEmitter, MexprParserListener.MexprParserListener):
    def __init__(self, mgr: AbstractVariableManager):
        super().__init__(mgr)
        self.mgr = mgr

    def exitMatchExprPlaceholder(self, ctx: MexprParser.MatchExprPlaceholderContext):
        param_defs: List[MexprParser.MexprPlaceholderParamContext] = ctx.mexprPlaceholderParam()
        params: List[language.Variable] = []
        for param_def in param_defs:
            param_id = parse_tree_text(param_def.ID())
            if param_def.varType() is not None:
                params.append(self.mgr.bv(param_id, parse_tree_text(param_def.varType())))
            else:
                params.append(self.mgr.bv(param_id, NONTERMINAL_PLACEHOLDER))

        name = self.mgr.fresh_name("mexprPlaceholder")
        mexpr_placeholder = MexprPlaceholderVariable(name, tuple(params))
        self.mgr.variables.setdefault(name, mexpr_placeholder)
        self.result.append(mexpr_placeholder)


ISLEARN_STANDARD_SEMANTIC_PREDICATES = STANDARD_SEMANTIC_PREDICATES | {
    INTERNET_CHECKSUM_PREDICATE,
    HEX_TO_DEC_PREDICATE,
}


def parse_abstract_isla(
        inp: str,
        grammar: Optional[Grammar] = None,
        structural_predicates: Set[StructuralPredicate] = STANDARD_STRUCTURAL_PREDICATES,
        semantic_predicates: Set[SemanticPredicate] = ISLEARN_STANDARD_SEMANTIC_PREDICATES) -> Formula:
    class BailPrintErrorStrategy(antlr4.BailErrorStrategy):
        def recover(self, recognizer: antlr4.Parser, e: antlr4.RecognitionException):
            recognizer._errHandler.reportError(recognizer, e)
            super().recover(recognizer, e)

    lexer = IslaLanguageLexer(InputStream(inp))
    parser = IslaLanguageParser(antlr4.CommonTokenStream(lexer))
    parser._errHandler = BailPrintErrorStrategy()
    isla_emitter = AbstractISLaEmitter(grammar, structural_predicates, semantic_predicates)
    antlr4.ParseTreeWalker().walk(isla_emitter, parser.start())
    return isla_emitter.result


class AbstractISLaUnparser(ISLaUnparser):
    def __init__(self, formula: Formula, indent="  "):
        super().__init__(formula, indent)

    def _unparse_smt_formula(self, formula: language.SMTFormula):
        result = smt_expr_to_str(formula.formula)

        for variable in formula.free_variables():
            if isinstance(variable, StringPlaceholderVariableTypes):
                result = result.replace(variable.name, str(variable))

        return [result]



import logging
from typing import Callable, Set

from grammar_graph import gg
from isla.fuzzer import GrammarFuzzer
from isla.helpers import is_nonterminal
from isla.language import DerivationTree
from isla.type_defs import Grammar, Path
from pathos import multiprocessing as pmp


class InputReducer:
    def __init__(self, grammar: Grammar, property: Callable[[DerivationTree], bool], k: int = 3):
        self.grammar = grammar
        self.graph = gg.GrammarGraph.from_grammar(grammar)
        self.property = property
        self.logger = logging.getLogger(__name__)
        self.k = k

    def reduce_by_smallest_subtree_replacement(self, inp: DerivationTree) -> DerivationTree:
        self.logger.debug("Reducing %s", inp)

        k_paths_in_inp = {
            path for path in inp.k_paths(self.graph, k=self.k)
            if not isinstance(path[-1], gg.TerminalNode) or
               (not isinstance(path[-1], gg.TerminalNode) and len(path[-1].symbol) > 1)}

        result = inp
        while True:
            def substitute_action(path: Path, _: DerivationTree):
                nonlocal result

                if not result.is_valid_path(path):
                    return

                subtree = result.get_subtree(path)

                matches = [t for p, t in subtree.filter(lambda t: t.value == subtree.value) if p]
                if not matches:
                    return

                matches = sorted(matches, key=DerivationTree.__len__)

                for match in matches:
                    # Don't replace subtree by match if it contains a k-path that will vanish otherwise
                    potential_replacement = result.replace_path(path, match)

                    k_paths_in_replacement = {
                        path for path in potential_replacement.k_paths(self.graph, k=self.k)
                        if (not isinstance(path[-1], gg.TerminalNode) or
                            (not isinstance(path[-1], gg.TerminalNode) and len(path[-1].symbol) > 1))}

                    if any(path not in k_paths_in_replacement for path in k_paths_in_inp):
                        continue

                    # self.logger.debug("Replacing %s with %s", subtree, match)

                    if self.property(potential_replacement):
                        result = potential_replacement
                        return

            # inp.traverse(substitute_action, kind=DerivationTree.TRAVERSE_PREORDER)
            inp.bfs(substitute_action)

            if result == inp:
                break

            inp = result

        self.logger.debug("Result: %s", result)
        return result

    def reduce_by_abstraction(self, inp: DerivationTree) -> DerivationTree:
        def can_abstract(tree: DerivationTree, path: Path) -> bool:
            subtree = tree.get_subtree(path)

            assert is_nonterminal(subtree.value)

            # for _ in range(10):
            #     new_leaf_tree = fuzzer.expand_tree(DerivationTree(subtree.value, None))
            #     new_tree = tree.replace_path(path, new_leaf_tree)
            #     if not self.property(new_tree):
            #         return False
            #
            # return True

            def check(_):
                new_leaf_tree = GrammarFuzzer(self.grammar).expand_tree(DerivationTree(subtree.value, None))
                new_tree = tree.replace_path(path, new_leaf_tree)
                if not self.property(new_tree):
                    return False
                return True

            with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
                for result in pool.uimap(check, range(10)):
                    if not result:
                        return False

                return True

        def reduce_action(path: Path, subtree: DerivationTree):
            nonlocal result
            if not is_nonterminal(subtree.value):
                return

            if path in do_not_abstract:
                return

            if can_abstract(result, path):
                result = result.replace_path(path, DerivationTree(subtree.value, None))
                return
            else:
                do_not_abstract.update({path[:idx] for idx in range(len(path))})

        result = inp
        do_not_abstract: Set[Path] = set([])
        result.traverse(reduce_action, kind=DerivationTree.TRAVERSE_POSTORDER)

        return result



# Generated from MexprParser.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .MexprParser import MexprParser
else:
    from MexprParser import MexprParser

# This class defines a complete listener for a parse tree produced by MexprParser.
class MexprParserListener(ParseTreeListener):

    # Enter a parse tree produced by MexprParser#matchExpr.
    def enterMatchExpr(self, ctx:MexprParser.MatchExprContext):
        pass

    # Exit a parse tree produced by MexprParser#matchExpr.
    def exitMatchExpr(self, ctx:MexprParser.MatchExprContext):
        pass


    # Enter a parse tree produced by MexprParser#MatchExprVar.
    def enterMatchExprVar(self, ctx:MexprParser.MatchExprVarContext):
        pass

    # Exit a parse tree produced by MexprParser#MatchExprVar.
    def exitMatchExprVar(self, ctx:MexprParser.MatchExprVarContext):
        pass


    # Enter a parse tree produced by MexprParser#MatchExprPlaceholder.
    def enterMatchExprPlaceholder(self, ctx:MexprParser.MatchExprPlaceholderContext):
        pass

    # Exit a parse tree produced by MexprParser#MatchExprPlaceholder.
    def exitMatchExprPlaceholder(self, ctx:MexprParser.MatchExprPlaceholderContext):
        pass


    # Enter a parse tree produced by MexprParser#MatchExprOptional.
    def enterMatchExprOptional(self, ctx:MexprParser.MatchExprOptionalContext):
        pass

    # Exit a parse tree produced by MexprParser#MatchExprOptional.
    def exitMatchExprOptional(self, ctx:MexprParser.MatchExprOptionalContext):
        pass


    # Enter a parse tree produced by MexprParser#MatchExprChars.
    def enterMatchExprChars(self, ctx:MexprParser.MatchExprCharsContext):
        pass

    # Exit a parse tree produced by MexprParser#MatchExprChars.
    def exitMatchExprChars(self, ctx:MexprParser.MatchExprCharsContext):
        pass


    # Enter a parse tree produced by MexprParser#mexprPlaceholderParam.
    def enterMexprPlaceholderParam(self, ctx:MexprParser.MexprPlaceholderParamContext):
        pass

    # Exit a parse tree produced by MexprParser#mexprPlaceholderParam.
    def exitMexprPlaceholderParam(self, ctx:MexprParser.MexprPlaceholderParamContext):
        pass


    # Enter a parse tree produced by MexprParser#varType.
    def enterVarType(self, ctx:MexprParser.VarTypeContext):
        pass

    # Exit a parse tree produced by MexprParser#varType.
    def exitVarType(self, ctx:MexprParser.VarTypeContext):
        pass



del MexprParser


import logging
import math
from typing import List, Callable, Tuple, Generator, Optional, Any, Sequence, TypeVar, Dict, cast

import datrie
from isla.helpers import is_nonterminal, mk_subtree_trie, path_to_trie_key
from isla.type_defs import Path, CanonicalGrammar

T = TypeVar("T")
S = TypeVar("S")
Tree = Tuple[T, Optional[List['Tree[T]']]]
DictTree = Dict[T, 'DictTree']


def dict_tree_from_paths(paths: Sequence[Sequence[T]]) -> DictTree[S]:
    def add_to_tree(root: DictTree, branch: Sequence[T]):
        if not branch:
            return
        if branch[0] not in root:
            root[branch[0]] = {}
        add_to_tree(root[branch[0]], branch[1:])

    tree: DictTree = {}
    for path in paths:
        add_to_tree(tree, path)

    assert len(tree) == 1  # Rooted tree

    return tree


def tree_from_paths(paths: Sequence[Sequence[T]]) -> Tree[S]:
    return dict_tree_to_tree(dict_tree_from_paths(paths))


def dict_tree_to_tree(tree: DictTree[T]) -> Tree[T]:
    assert len(tree) == 1  # Rooted tree
    label: T = next(iter(tree))
    return label, [dict_tree_to_tree(child) for child in [{t[0]: t[1]} for t in tree[label].items()]]


def get_subtree(tree: Tree[T], path: Path) -> Tree[T]:
    """Access a subtree based on `path` (a list of children numbers)"""
    curr_node = tree
    while path:
        curr_node = curr_node[1][path[0]]
        path = path[1:]

    return curr_node


def replace_path(
        tree: Tree[T],
        path: Path,
        replacement_tree: Tree[T]) -> Tree[T]:
    """Returns tree where replacement_tree has been inserted at `path` instead of the original subtree"""
    stack: List[Tree[T]] = [tree]
    for path_elem in path:
        stack.append(stack[-1][1][path_elem])

    stack[-1] = replacement_tree

    for path_elem in reversed(path):
        assert len(stack) > 1
        replacement = stack.pop()
        parent = stack.pop()

        node, children = parent
        new_children = children[:path_elem] + [replacement] + children[path_elem + 1:]

        stack.append((node, new_children))

    assert len(stack) == 1
    return stack[0]


def filter_tree(
        tree: Tree,
        f: Callable[[Tree], bool],
        enforce_unique: bool = False) -> List[Tuple[Path, Tree]]:
    result: List[Tuple[Path, Tree]] = []

    for path, subtree in tree_paths(tree):
        if f(subtree):
            result.append((path, subtree))

            if enforce_unique and len(result) > 1:
                raise RuntimeError(f"Found searched-for element more than once in {tree}")

    return result


TRAVERSE_PREORDER = 0
TRAVERSE_POSTORDER = 1


def traverse_tree(
        tree: Tree[T],
        action: Callable[[Path, Tree[T]], None],
        abort_condition: Callable[[Path, Tree[T]], bool] = lambda p, n: False,
        kind: int = TRAVERSE_PREORDER,
        reverse: bool = False) -> None:
    stack_1: List[Tuple[Path, Tree[T]]] = [((), tree)]
    stack_2: List[Tuple[Path, Tree[T]]] = []

    if kind == TRAVERSE_PREORDER:
        reverse = not reverse

    while stack_1:
        path, node = stack_1.pop()

        if abort_condition(path, node):
            return

        if kind == TRAVERSE_POSTORDER:
            stack_2.append((path, node))

        if kind == TRAVERSE_PREORDER:
            action(path, node)

        if node[1]:
            iterator = reversed(node[1]) if reverse else iter(node[1])

            for idx, child in enumerate(iterator):
                new_path = path + ((len(node[1]) - idx - 1) if reverse else idx,)
                stack_1.append((new_path, child))

    if kind == TRAVERSE_POSTORDER:
        while stack_2:
            action(*stack_2.pop())


def tree_paths(tree: Tree[T]) -> List[Tuple[Path, Tree[T]]]:
    def action(path, node):
        result.append((path, node))

    result: List[Tuple[Path, Tree[T]]] = []
    traverse_tree(tree, action, kind=TRAVERSE_PREORDER)
    return result


def trie_from_parse_tree(tree: Tree) -> datrie.Trie:
    pass
    trie = mk_subtree_trie()  # Works for up to 30 children of a node
    for path, subtree in tree_paths(tree):
        trie[path_to_trie_key(path)] = (path, subtree)
    return trie


def next_trie_key(trie: datrie.Trie, path: Path | str) -> Optional[str]:
    if isinstance(path, tuple):
        path = path_to_trie_key(path)

    suffixes = tuple(filter(None, trie.suffixes(path)))
    if suffixes:
        return path + suffixes[0]
    else:
        prefixes = reversed(trie.prefixes(path))
        for prefix in prefixes:
            maybe_next_key = prefix[:-1] + chr(ord(prefix[-1]) + 1)
            if trie.keys(maybe_next_key):
                return maybe_next_key

        return None


def tree_to_string(tree: Tree, show_open_leaves: bool = False) -> str:
    result = []
    stack = [tree]

    while stack:
        node = stack.pop(0)
        symbol, children = node

        if not children:
            if children is not None:
                result.append("" if is_nonterminal(symbol) else symbol)
            else:
                result.append(symbol if show_open_leaves else "")

            continue

        stack = list(children) + stack

    return ''.join(result)


def open_leaves(tree: Tree) -> Generator[Tuple[Path, Tree], None, None]:
    return ((path, sub_tree) for path, sub_tree in tree_paths(tree) if sub_tree[1] is None)


def tree_leaves(tree: Tree) -> Generator[Tuple[Path, Tree], None, None]:
    return ((path, sub_tree) for path, sub_tree in tree_paths(tree) if not sub_tree[1])


def expand_tree(
        tree: Tree,
        canonical_grammar: CanonicalGrammar,
        limit: Optional[int] = None,
        expand_beyond_nonterminals: bool = False) -> List[Tree]:
    nonterminal_expansions = {
        (leaf_path, leaf_node): [
            [(child, None if is_nonterminal(child) else [])
             for child in expansion]
            for expansion in canonical_grammar[leaf_node[0]]
            if expand_beyond_nonterminals or any(is_nonterminal(child) for child in expansion)
        ]
        for leaf_path, leaf_node in open_leaves(tree)
    }

    if all(not expansion for expansion in nonterminal_expansions.values()):
        return []

    result: List[Tree] = [tree]
    for (leaf_path, leaf_node), expansions in nonterminal_expansions.items():
        previous_result = result
        result = []
        for t in previous_result:
            if not expansions:
                result.append(t)

            for expansion in expansions:
                result.append(replace_path(
                    t,
                    leaf_path,
                    (leaf_node[0], expansion)))

                if limit and len(result) >= limit:
                    break
            if limit and len(result) >= limit:
                break
        if limit and len(result) >= limit:
            break

    # assert ((limit and len(result) == limit) or
    #         len(result) == math.prod(len(values) for values in nonterminal_expansions.values()))

    return result



# Generated from IslaLanguage.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,46,185,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,1,0,3,0,12,8,
        0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,2,1,2,1,2,3,2,26,8,2,1,2,
        1,2,1,2,3,2,31,8,2,3,2,33,8,2,1,2,1,2,1,2,1,2,1,2,3,2,40,8,2,1,2,
        1,2,1,2,3,2,45,8,2,3,2,47,8,2,1,2,1,2,1,2,1,2,1,2,3,2,54,8,2,1,2,
        1,2,1,2,1,2,1,2,3,2,61,8,2,3,2,63,8,2,1,2,1,2,1,2,1,2,1,2,3,2,70,
        8,2,1,2,1,2,1,2,1,2,1,2,3,2,77,8,2,3,2,79,8,2,1,2,1,2,1,2,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,5,2,100,
        8,2,10,2,12,2,103,9,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,3,2,112,8,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,5,2,129,
        8,2,10,2,12,2,132,9,2,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,
        3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,4,3,158,8,
        3,11,3,12,3,159,1,3,1,3,3,3,164,8,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,
        1,3,1,3,1,3,1,3,1,3,5,3,178,8,3,10,3,12,3,181,9,3,1,4,1,4,1,4,0,
        2,4,6,5,0,2,4,6,8,0,8,2,0,25,25,28,28,3,0,7,7,20,23,37,44,1,0,20,
        21,1,0,22,23,1,0,39,40,1,0,37,38,2,0,7,7,41,44,1,0,24,31,225,0,11,
        1,0,0,0,2,15,1,0,0,0,4,111,1,0,0,0,6,163,1,0,0,0,8,182,1,0,0,0,10,
        12,3,2,1,0,11,10,1,0,0,0,11,12,1,0,0,0,12,13,1,0,0,0,13,14,3,4,2,
        0,14,1,1,0,0,0,15,16,5,1,0,0,16,17,5,30,0,0,17,18,5,2,0,0,18,19,
        5,28,0,0,19,20,5,3,0,0,20,3,1,0,0,0,21,22,6,2,-1,0,22,23,5,4,0,0,
        23,25,7,0,0,0,24,26,5,30,0,0,25,24,1,0,0,0,25,26,1,0,0,0,26,32,1,
        0,0,0,27,30,5,5,0,0,28,31,5,30,0,0,29,31,5,28,0,0,30,28,1,0,0,0,
        30,29,1,0,0,0,31,33,1,0,0,0,32,27,1,0,0,0,32,33,1,0,0,0,33,34,1,
        0,0,0,34,35,5,2,0,0,35,112,3,4,2,15,36,37,5,6,0,0,37,39,7,0,0,0,
        38,40,5,30,0,0,39,38,1,0,0,0,39,40,1,0,0,0,40,46,1,0,0,0,41,44,5,
        5,0,0,42,45,5,30,0,0,43,45,5,28,0,0,44,42,1,0,0,0,44,43,1,0,0,0,
        45,47,1,0,0,0,46,41,1,0,0,0,46,47,1,0,0,0,47,48,1,0,0,0,48,49,5,
        2,0,0,49,112,3,4,2,14,50,51,5,4,0,0,51,53,7,0,0,0,52,54,5,30,0,0,
        53,52,1,0,0,0,53,54,1,0,0,0,54,55,1,0,0,0,55,56,5,7,0,0,56,62,5,
        29,0,0,57,60,5,5,0,0,58,61,5,30,0,0,59,61,5,28,0,0,60,58,1,0,0,0,
        60,59,1,0,0,0,61,63,1,0,0,0,62,57,1,0,0,0,62,63,1,0,0,0,63,64,1,
        0,0,0,64,65,5,2,0,0,65,112,3,4,2,13,66,67,5,6,0,0,67,69,7,0,0,0,
        68,70,5,30,0,0,69,68,1,0,0,0,69,70,1,0,0,0,70,71,1,0,0,0,71,72,5,
        7,0,0,72,78,5,29,0,0,73,76,5,5,0,0,74,77,5,30,0,0,75,77,5,28,0,0,
        76,74,1,0,0,0,76,75,1,0,0,0,77,79,1,0,0,0,78,73,1,0,0,0,78,79,1,
        0,0,0,79,80,1,0,0,0,80,81,5,2,0,0,81,112,3,4,2,12,82,83,5,6,0,0,
        83,84,5,8,0,0,84,85,5,30,0,0,85,86,5,2,0,0,86,112,3,4,2,11,87,88,
        5,4,0,0,88,89,5,8,0,0,89,90,5,30,0,0,90,91,5,2,0,0,91,112,3,4,2,
        10,92,93,5,9,0,0,93,112,3,4,2,9,94,95,5,30,0,0,95,96,5,15,0,0,96,
        101,3,8,4,0,97,98,5,16,0,0,98,100,3,8,4,0,99,97,1,0,0,0,100,103,
        1,0,0,0,101,99,1,0,0,0,101,102,1,0,0,0,102,104,1,0,0,0,103,101,1,
        0,0,0,104,105,5,17,0,0,105,112,1,0,0,0,106,112,3,6,3,0,107,108,5,
        15,0,0,108,109,3,4,2,0,109,110,5,17,0,0,110,112,1,0,0,0,111,21,1,
        0,0,0,111,36,1,0,0,0,111,50,1,0,0,0,111,66,1,0,0,0,111,82,1,0,0,
        0,111,87,1,0,0,0,111,92,1,0,0,0,111,94,1,0,0,0,111,106,1,0,0,0,111,
        107,1,0,0,0,112,130,1,0,0,0,113,114,10,8,0,0,114,115,5,10,0,0,115,
        129,3,4,2,9,116,117,10,7,0,0,117,118,5,11,0,0,118,129,3,4,2,8,119,
        120,10,6,0,0,120,121,5,12,0,0,121,129,3,4,2,7,122,123,10,5,0,0,123,
        124,5,13,0,0,124,129,3,4,2,6,125,126,10,4,0,0,126,127,5,14,0,0,127,
        129,3,4,2,5,128,113,1,0,0,0,128,116,1,0,0,0,128,119,1,0,0,0,128,
        122,1,0,0,0,128,125,1,0,0,0,129,132,1,0,0,0,130,128,1,0,0,0,130,
        131,1,0,0,0,131,5,1,0,0,0,132,130,1,0,0,0,133,134,6,3,-1,0,134,164,
        5,18,0,0,135,164,5,19,0,0,136,164,5,31,0,0,137,164,5,30,0,0,138,
        164,5,24,0,0,139,164,5,28,0,0,140,164,5,29,0,0,141,164,5,25,0,0,
        142,164,5,26,0,0,143,164,5,27,0,0,144,164,7,1,0,0,145,146,7,2,0,
        0,146,147,5,15,0,0,147,148,3,6,3,0,148,149,5,17,0,0,149,164,1,0,
        0,0,150,151,5,15,0,0,151,152,3,6,3,0,152,153,5,17,0,0,153,164,1,
        0,0,0,154,155,5,15,0,0,155,157,3,6,3,0,156,158,3,6,3,0,157,156,1,
        0,0,0,158,159,1,0,0,0,159,157,1,0,0,0,159,160,1,0,0,0,160,161,1,
        0,0,0,161,162,5,17,0,0,162,164,1,0,0,0,163,133,1,0,0,0,163,135,1,
        0,0,0,163,136,1,0,0,0,163,137,1,0,0,0,163,138,1,0,0,0,163,139,1,
        0,0,0,163,140,1,0,0,0,163,141,1,0,0,0,163,142,1,0,0,0,163,143,1,
        0,0,0,163,144,1,0,0,0,163,145,1,0,0,0,163,150,1,0,0,0,163,154,1,
        0,0,0,164,179,1,0,0,0,165,166,10,6,0,0,166,167,7,3,0,0,167,178,3,
        6,3,7,168,169,10,5,0,0,169,170,7,4,0,0,170,178,3,6,3,6,171,172,10,
        4,0,0,172,173,7,5,0,0,173,178,3,6,3,5,174,175,10,3,0,0,175,176,7,
        6,0,0,176,178,3,6,3,4,177,165,1,0,0,0,177,168,1,0,0,0,177,171,1,
        0,0,0,177,174,1,0,0,0,178,181,1,0,0,0,179,177,1,0,0,0,179,180,1,
        0,0,0,180,7,1,0,0,0,181,179,1,0,0,0,182,183,7,7,0,0,183,9,1,0,0,
        0,21,11,25,30,32,39,44,46,53,60,62,69,76,78,101,111,128,130,159,
        163,177,179
    ]

class IslaLanguageParser ( Parser ):

    grammarFileName = "IslaLanguage.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'const'", "':'", "';'", "'forall'", "'in'", 
                     "'exists'", "'='", "'int'", "'not'", "'and'", "'or'", 
                     "'xor'", "'implies'", "'iff'", "'('", "','", "')'", 
                     "'true'", "'false'", "'re.+'", "'re.*'", "'re.++'", 
                     "'str.++'", "<INVALID>", "'<?NONTERMINAL>'", "'<?STRING>'", 
                     "'<?DSTRINGS>'", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'.'", "'..'", "'['", "']'", 
                     "'/'", "'*'", "'+'", "'-'", "'>='", "'<='", "'>'", 
                     "'<'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "XPATHEXPR", "NONTERMINAL_PH", "STRING_PH", "DSTRINGS_PH", 
                      "VAR_TYPE", "STRING", "ID", "INT", "ESC", "DOT", "TWODOTS", 
                      "BROP", "BRCL", "DIV", "MUL", "PLUS", "MINUS", "GEQ", 
                      "LEQ", "GT", "LT", "WS", "LINE_COMMENT" ]

    RULE_start = 0
    RULE_constDecl = 1
    RULE_formula = 2
    RULE_sexpr = 3
    RULE_predicateArg = 4

    ruleNames =  [ "start", "constDecl", "formula", "sexpr", "predicateArg" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    XPATHEXPR=24
    NONTERMINAL_PH=25
    STRING_PH=26
    DSTRINGS_PH=27
    VAR_TYPE=28
    STRING=29
    ID=30
    INT=31
    ESC=32
    DOT=33
    TWODOTS=34
    BROP=35
    BRCL=36
    DIV=37
    MUL=38
    PLUS=39
    MINUS=40
    GEQ=41
    LEQ=42
    GT=43
    LT=44
    WS=45
    LINE_COMMENT=46

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)


        def constDecl(self):
            return self.getTypedRuleContext(IslaLanguageParser.ConstDeclContext,0)


        def getRuleIndex(self):
            return IslaLanguageParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)




    def start(self):

        localctx = IslaLanguageParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 11
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==IslaLanguageParser.T__0:
                self.state = 10
                self.constDecl()


            self.state = 13
            self.formula(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ConstDeclContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)

        def VAR_TYPE(self):
            return self.getToken(IslaLanguageParser.VAR_TYPE, 0)

        def getRuleIndex(self):
            return IslaLanguageParser.RULE_constDecl

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterConstDecl" ):
                listener.enterConstDecl(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitConstDecl" ):
                listener.exitConstDecl(self)




    def constDecl(self):

        localctx = IslaLanguageParser.ConstDeclContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_constDecl)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self.match(IslaLanguageParser.T__0)
            self.state = 16
            self.match(IslaLanguageParser.ID)
            self.state = 17
            self.match(IslaLanguageParser.T__1)
            self.state = 18
            self.match(IslaLanguageParser.VAR_TYPE)
            self.state = 19
            self.match(IslaLanguageParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FormulaContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return IslaLanguageParser.RULE_formula

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class ExistsMexprContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.boundVarType = None # Token
            self.varId = None # Token
            self.inId = None # Token
            self.inVarType = None # Token
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(IslaLanguageParser.STRING, 0)
        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)

        def VAR_TYPE(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.VAR_TYPE)
            else:
                return self.getToken(IslaLanguageParser.VAR_TYPE, i)
        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)
        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.ID)
            else:
                return self.getToken(IslaLanguageParser.ID, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExistsMexpr" ):
                listener.enterExistsMexpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExistsMexpr" ):
                listener.exitExistsMexpr(self)


    class NegationContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNegation" ):
                listener.enterNegation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNegation" ):
                listener.exitNegation(self)


    class ImplicationContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.FormulaContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterImplication" ):
                listener.enterImplication(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitImplication" ):
                listener.exitImplication(self)


    class ForallMexprContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.boundVarType = None # Token
            self.varId = None # Token
            self.inId = None # Token
            self.inVarType = None # Token
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(IslaLanguageParser.STRING, 0)
        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)

        def VAR_TYPE(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.VAR_TYPE)
            else:
                return self.getToken(IslaLanguageParser.VAR_TYPE, i)
        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)
        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.ID)
            else:
                return self.getToken(IslaLanguageParser.ID, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterForallMexpr" ):
                listener.enterForallMexpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitForallMexpr" ):
                listener.exitForallMexpr(self)


    class ExistsIntContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)
        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExistsInt" ):
                listener.enterExistsInt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExistsInt" ):
                listener.exitExistsInt(self)


    class DisjunctionContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.FormulaContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDisjunction" ):
                listener.enterDisjunction(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDisjunction" ):
                listener.exitDisjunction(self)


    class PredicateAtomContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)
        def predicateArg(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.PredicateArgContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.PredicateArgContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicateAtom" ):
                listener.enterPredicateAtom(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicateAtom" ):
                listener.exitPredicateAtom(self)


    class SMTFormulaContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def sexpr(self):
            return self.getTypedRuleContext(IslaLanguageParser.SexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSMTFormula" ):
                listener.enterSMTFormula(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSMTFormula" ):
                listener.exitSMTFormula(self)


    class EquivalenceContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.FormulaContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEquivalence" ):
                listener.enterEquivalence(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEquivalence" ):
                listener.exitEquivalence(self)


    class ExistsContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.boundVarType = None # Token
            self.varId = None # Token
            self.inId = None # Token
            self.inVarType = None # Token
            self.copyFrom(ctx)

        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)

        def VAR_TYPE(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.VAR_TYPE)
            else:
                return self.getToken(IslaLanguageParser.VAR_TYPE, i)
        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)
        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.ID)
            else:
                return self.getToken(IslaLanguageParser.ID, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExists" ):
                listener.enterExists(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExists" ):
                listener.exitExists(self)


    class ConjunctionContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.FormulaContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterConjunction" ):
                listener.enterConjunction(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitConjunction" ):
                listener.exitConjunction(self)


    class ParFormulaContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterParFormula" ):
                listener.enterParFormula(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitParFormula" ):
                listener.exitParFormula(self)


    class ForallContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.boundVarType = None # Token
            self.varId = None # Token
            self.inId = None # Token
            self.inVarType = None # Token
            self.copyFrom(ctx)

        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)

        def VAR_TYPE(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.VAR_TYPE)
            else:
                return self.getToken(IslaLanguageParser.VAR_TYPE, i)
        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)
        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(IslaLanguageParser.ID)
            else:
                return self.getToken(IslaLanguageParser.ID, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterForall" ):
                listener.enterForall(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitForall" ):
                listener.exitForall(self)


    class ExclusiveOrContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def formula(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.FormulaContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExclusiveOr" ):
                listener.enterExclusiveOr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExclusiveOr" ):
                listener.exitExclusiveOr(self)


    class ForallIntContext(FormulaContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.FormulaContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)
        def formula(self):
            return self.getTypedRuleContext(IslaLanguageParser.FormulaContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterForallInt" ):
                listener.enterForallInt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitForallInt" ):
                listener.exitForallInt(self)



    def formula(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = IslaLanguageParser.FormulaContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(localctx, 4, self.RULE_formula, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 111
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,14,self._ctx)
            if la_ == 1:
                localctx = IslaLanguageParser.ForallContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 22
                self.match(IslaLanguageParser.T__3)

                self.state = 23
                localctx.boundVarType = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==IslaLanguageParser.NONTERMINAL_PH or _la==IslaLanguageParser.VAR_TYPE):
                    localctx.boundVarType = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 25
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.ID:
                    self.state = 24
                    localctx.varId = self.match(IslaLanguageParser.ID)


                self.state = 32
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.T__4:
                    self.state = 27
                    self.match(IslaLanguageParser.T__4)
                    self.state = 30
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [IslaLanguageParser.ID]:
                        self.state = 28
                        localctx.inId = self.match(IslaLanguageParser.ID)
                        pass
                    elif token in [IslaLanguageParser.VAR_TYPE]:
                        self.state = 29
                        localctx.inVarType = self.match(IslaLanguageParser.VAR_TYPE)
                        pass
                    else:
                        raise NoViableAltException(self)



                self.state = 34
                self.match(IslaLanguageParser.T__1)
                self.state = 35
                self.formula(15)
                pass

            elif la_ == 2:
                localctx = IslaLanguageParser.ExistsContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 36
                self.match(IslaLanguageParser.T__5)

                self.state = 37
                localctx.boundVarType = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==IslaLanguageParser.NONTERMINAL_PH or _la==IslaLanguageParser.VAR_TYPE):
                    localctx.boundVarType = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 39
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.ID:
                    self.state = 38
                    localctx.varId = self.match(IslaLanguageParser.ID)


                self.state = 46
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.T__4:
                    self.state = 41
                    self.match(IslaLanguageParser.T__4)
                    self.state = 44
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [IslaLanguageParser.ID]:
                        self.state = 42
                        localctx.inId = self.match(IslaLanguageParser.ID)
                        pass
                    elif token in [IslaLanguageParser.VAR_TYPE]:
                        self.state = 43
                        localctx.inVarType = self.match(IslaLanguageParser.VAR_TYPE)
                        pass
                    else:
                        raise NoViableAltException(self)



                self.state = 48
                self.match(IslaLanguageParser.T__1)
                self.state = 49
                self.formula(14)
                pass

            elif la_ == 3:
                localctx = IslaLanguageParser.ForallMexprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 50
                self.match(IslaLanguageParser.T__3)

                self.state = 51
                localctx.boundVarType = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==IslaLanguageParser.NONTERMINAL_PH or _la==IslaLanguageParser.VAR_TYPE):
                    localctx.boundVarType = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 53
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.ID:
                    self.state = 52
                    localctx.varId = self.match(IslaLanguageParser.ID)


                self.state = 55
                self.match(IslaLanguageParser.T__6)
                self.state = 56
                self.match(IslaLanguageParser.STRING)
                self.state = 62
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.T__4:
                    self.state = 57
                    self.match(IslaLanguageParser.T__4)
                    self.state = 60
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [IslaLanguageParser.ID]:
                        self.state = 58
                        localctx.inId = self.match(IslaLanguageParser.ID)
                        pass
                    elif token in [IslaLanguageParser.VAR_TYPE]:
                        self.state = 59
                        localctx.inVarType = self.match(IslaLanguageParser.VAR_TYPE)
                        pass
                    else:
                        raise NoViableAltException(self)



                self.state = 64
                self.match(IslaLanguageParser.T__1)
                self.state = 65
                self.formula(13)
                pass

            elif la_ == 4:
                localctx = IslaLanguageParser.ExistsMexprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 66
                self.match(IslaLanguageParser.T__5)

                self.state = 67
                localctx.boundVarType = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==IslaLanguageParser.NONTERMINAL_PH or _la==IslaLanguageParser.VAR_TYPE):
                    localctx.boundVarType = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 69
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.ID:
                    self.state = 68
                    localctx.varId = self.match(IslaLanguageParser.ID)


                self.state = 71
                self.match(IslaLanguageParser.T__6)
                self.state = 72
                self.match(IslaLanguageParser.STRING)
                self.state = 78
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==IslaLanguageParser.T__4:
                    self.state = 73
                    self.match(IslaLanguageParser.T__4)
                    self.state = 76
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [IslaLanguageParser.ID]:
                        self.state = 74
                        localctx.inId = self.match(IslaLanguageParser.ID)
                        pass
                    elif token in [IslaLanguageParser.VAR_TYPE]:
                        self.state = 75
                        localctx.inVarType = self.match(IslaLanguageParser.VAR_TYPE)
                        pass
                    else:
                        raise NoViableAltException(self)



                self.state = 80
                self.match(IslaLanguageParser.T__1)
                self.state = 81
                self.formula(12)
                pass

            elif la_ == 5:
                localctx = IslaLanguageParser.ExistsIntContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 82
                self.match(IslaLanguageParser.T__5)
                self.state = 83
                self.match(IslaLanguageParser.T__7)
                self.state = 84
                self.match(IslaLanguageParser.ID)
                self.state = 85
                self.match(IslaLanguageParser.T__1)
                self.state = 86
                self.formula(11)
                pass

            elif la_ == 6:
                localctx = IslaLanguageParser.ForallIntContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 87
                self.match(IslaLanguageParser.T__3)
                self.state = 88
                self.match(IslaLanguageParser.T__7)
                self.state = 89
                self.match(IslaLanguageParser.ID)
                self.state = 90
                self.match(IslaLanguageParser.T__1)
                self.state = 91
                self.formula(10)
                pass

            elif la_ == 7:
                localctx = IslaLanguageParser.NegationContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 92
                self.match(IslaLanguageParser.T__8)
                self.state = 93
                self.formula(9)
                pass

            elif la_ == 8:
                localctx = IslaLanguageParser.PredicateAtomContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 94
                self.match(IslaLanguageParser.ID)
                self.state = 95
                self.match(IslaLanguageParser.T__14)
                self.state = 96
                self.predicateArg()
                self.state = 101
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==IslaLanguageParser.T__15:
                    self.state = 97
                    self.match(IslaLanguageParser.T__15)
                    self.state = 98
                    self.predicateArg()
                    self.state = 103
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 104
                self.match(IslaLanguageParser.T__16)
                pass

            elif la_ == 9:
                localctx = IslaLanguageParser.SMTFormulaContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 106
                self.sexpr(0)
                pass

            elif la_ == 10:
                localctx = IslaLanguageParser.ParFormulaContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 107
                self.match(IslaLanguageParser.T__14)
                self.state = 108
                self.formula(0)
                self.state = 109
                self.match(IslaLanguageParser.T__16)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 130
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,16,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 128
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,15,self._ctx)
                    if la_ == 1:
                        localctx = IslaLanguageParser.ConjunctionContext(self, IslaLanguageParser.FormulaContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_formula)
                        self.state = 113
                        if not self.precpred(self._ctx, 8):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 8)")
                        self.state = 114
                        self.match(IslaLanguageParser.T__9)
                        self.state = 115
                        self.formula(9)
                        pass

                    elif la_ == 2:
                        localctx = IslaLanguageParser.DisjunctionContext(self, IslaLanguageParser.FormulaContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_formula)
                        self.state = 116
                        if not self.precpred(self._ctx, 7):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
                        self.state = 117
                        self.match(IslaLanguageParser.T__10)
                        self.state = 118
                        self.formula(8)
                        pass

                    elif la_ == 3:
                        localctx = IslaLanguageParser.ExclusiveOrContext(self, IslaLanguageParser.FormulaContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_formula)
                        self.state = 119
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
                        self.state = 120
                        self.match(IslaLanguageParser.T__11)
                        self.state = 121
                        self.formula(7)
                        pass

                    elif la_ == 4:
                        localctx = IslaLanguageParser.ImplicationContext(self, IslaLanguageParser.FormulaContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_formula)
                        self.state = 122
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 123
                        self.match(IslaLanguageParser.T__12)
                        self.state = 124
                        self.formula(6)
                        pass

                    elif la_ == 5:
                        localctx = IslaLanguageParser.EquivalenceContext(self, IslaLanguageParser.FormulaContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_formula)
                        self.state = 125
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 126
                        self.match(IslaLanguageParser.T__13)
                        self.state = 127
                        self.formula(5)
                        pass

             
                self.state = 132
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class SexprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return IslaLanguageParser.RULE_sexpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class SexprNonterminalStringPhContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprNonterminalStringPh" ):
                listener.enterSexprNonterminalStringPh(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprNonterminalStringPh" ):
                listener.exitSexprNonterminalStringPh(self)


    class SexprInfixReStrContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def sexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.SexprContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.SexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprInfixReStr" ):
                listener.enterSexprInfixReStr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprInfixReStr" ):
                listener.exitSexprInfixReStr(self)


    class SexprNumContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INT(self):
            return self.getToken(IslaLanguageParser.INT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprNum" ):
                listener.enterSexprNum(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprNum" ):
                listener.exitSexprNum(self)


    class SexprXPathExprContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def XPATHEXPR(self):
            return self.getToken(IslaLanguageParser.XPATHEXPR, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprXPathExpr" ):
                listener.enterSexprXPathExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprXPathExpr" ):
                listener.exitSexprXPathExpr(self)


    class SexprOpContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def GEQ(self):
            return self.getToken(IslaLanguageParser.GEQ, 0)
        def LEQ(self):
            return self.getToken(IslaLanguageParser.LEQ, 0)
        def GT(self):
            return self.getToken(IslaLanguageParser.GT, 0)
        def LT(self):
            return self.getToken(IslaLanguageParser.LT, 0)
        def DIV(self):
            return self.getToken(IslaLanguageParser.DIV, 0)
        def MUL(self):
            return self.getToken(IslaLanguageParser.MUL, 0)
        def PLUS(self):
            return self.getToken(IslaLanguageParser.PLUS, 0)
        def MINUS(self):
            return self.getToken(IslaLanguageParser.MINUS, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprOp" ):
                listener.enterSexprOp(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprOp" ):
                listener.exitSexprOp(self)


    class SexprInfixPlusMinusContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def sexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.SexprContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.SexprContext,i)

        def PLUS(self):
            return self.getToken(IslaLanguageParser.PLUS, 0)
        def MINUS(self):
            return self.getToken(IslaLanguageParser.MINUS, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprInfixPlusMinus" ):
                listener.enterSexprInfixPlusMinus(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprInfixPlusMinus" ):
                listener.exitSexprInfixPlusMinus(self)


    class SexprTrueContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprTrue" ):
                listener.enterSexprTrue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprTrue" ):
                listener.exitSexprTrue(self)


    class SexprFreeIdContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def VAR_TYPE(self):
            return self.getToken(IslaLanguageParser.VAR_TYPE, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprFreeId" ):
                listener.enterSexprFreeId(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprFreeId" ):
                listener.exitSexprFreeId(self)


    class SexprStringPhContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STRING_PH(self):
            return self.getToken(IslaLanguageParser.STRING_PH, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprStringPh" ):
                listener.enterSexprStringPh(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprStringPh" ):
                listener.exitSexprStringPh(self)


    class SepxrAppContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # SexprContext
            self.copyFrom(ctx)

        def sexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.SexprContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.SexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSepxrApp" ):
                listener.enterSepxrApp(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSepxrApp" ):
                listener.exitSepxrApp(self)


    class SexprInfixMulDivContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def sexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.SexprContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.SexprContext,i)

        def MUL(self):
            return self.getToken(IslaLanguageParser.MUL, 0)
        def DIV(self):
            return self.getToken(IslaLanguageParser.DIV, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprInfixMulDiv" ):
                listener.enterSexprInfixMulDiv(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprInfixMulDiv" ):
                listener.exitSexprInfixMulDiv(self)


    class SexprIdContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprId" ):
                listener.enterSexprId(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprId" ):
                listener.exitSexprId(self)


    class SexprPrefixContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def sexpr(self):
            return self.getTypedRuleContext(IslaLanguageParser.SexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprPrefix" ):
                listener.enterSexprPrefix(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprPrefix" ):
                listener.exitSexprPrefix(self)


    class SepxrParenContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def sexpr(self):
            return self.getTypedRuleContext(IslaLanguageParser.SexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSepxrParen" ):
                listener.enterSepxrParen(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSepxrParen" ):
                listener.exitSepxrParen(self)


    class SexprStrContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STRING(self):
            return self.getToken(IslaLanguageParser.STRING, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprStr" ):
                listener.enterSexprStr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprStr" ):
                listener.exitSexprStr(self)


    class SexprInfixEqContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def sexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(IslaLanguageParser.SexprContext)
            else:
                return self.getTypedRuleContext(IslaLanguageParser.SexprContext,i)

        def GEQ(self):
            return self.getToken(IslaLanguageParser.GEQ, 0)
        def LEQ(self):
            return self.getToken(IslaLanguageParser.LEQ, 0)
        def GT(self):
            return self.getToken(IslaLanguageParser.GT, 0)
        def LT(self):
            return self.getToken(IslaLanguageParser.LT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprInfixEq" ):
                listener.enterSexprInfixEq(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprInfixEq" ):
                listener.exitSexprInfixEq(self)


    class SexprDisjStringsPhContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def DSTRINGS_PH(self):
            return self.getToken(IslaLanguageParser.DSTRINGS_PH, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprDisjStringsPh" ):
                listener.enterSexprDisjStringsPh(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprDisjStringsPh" ):
                listener.exitSexprDisjStringsPh(self)


    class SexprFalseContext(SexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a IslaLanguageParser.SexprContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSexprFalse" ):
                listener.enterSexprFalse(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSexprFalse" ):
                listener.exitSexprFalse(self)



    def sexpr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = IslaLanguageParser.SexprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 6
        self.enterRecursionRule(localctx, 6, self.RULE_sexpr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 163
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
            if la_ == 1:
                localctx = IslaLanguageParser.SexprTrueContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 134
                self.match(IslaLanguageParser.T__17)
                pass

            elif la_ == 2:
                localctx = IslaLanguageParser.SexprFalseContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 135
                self.match(IslaLanguageParser.T__18)
                pass

            elif la_ == 3:
                localctx = IslaLanguageParser.SexprNumContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 136
                self.match(IslaLanguageParser.INT)
                pass

            elif la_ == 4:
                localctx = IslaLanguageParser.SexprIdContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 137
                self.match(IslaLanguageParser.ID)
                pass

            elif la_ == 5:
                localctx = IslaLanguageParser.SexprXPathExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 138
                self.match(IslaLanguageParser.XPATHEXPR)
                pass

            elif la_ == 6:
                localctx = IslaLanguageParser.SexprFreeIdContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 139
                self.match(IslaLanguageParser.VAR_TYPE)
                pass

            elif la_ == 7:
                localctx = IslaLanguageParser.SexprStrContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 140
                self.match(IslaLanguageParser.STRING)
                pass

            elif la_ == 8:
                localctx = IslaLanguageParser.SexprNonterminalStringPhContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 141
                self.match(IslaLanguageParser.NONTERMINAL_PH)
                pass

            elif la_ == 9:
                localctx = IslaLanguageParser.SexprStringPhContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 142
                self.match(IslaLanguageParser.STRING_PH)
                pass

            elif la_ == 10:
                localctx = IslaLanguageParser.SexprDisjStringsPhContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 143
                self.match(IslaLanguageParser.DSTRINGS_PH)
                pass

            elif la_ == 11:
                localctx = IslaLanguageParser.SexprOpContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 144
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << IslaLanguageParser.T__6) | (1 << IslaLanguageParser.T__19) | (1 << IslaLanguageParser.T__20) | (1 << IslaLanguageParser.T__21) | (1 << IslaLanguageParser.T__22) | (1 << IslaLanguageParser.DIV) | (1 << IslaLanguageParser.MUL) | (1 << IslaLanguageParser.PLUS) | (1 << IslaLanguageParser.MINUS) | (1 << IslaLanguageParser.GEQ) | (1 << IslaLanguageParser.LEQ) | (1 << IslaLanguageParser.GT) | (1 << IslaLanguageParser.LT))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 12:
                localctx = IslaLanguageParser.SexprPrefixContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 145
                localctx.op = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==IslaLanguageParser.T__19 or _la==IslaLanguageParser.T__20):
                    localctx.op = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 146
                self.match(IslaLanguageParser.T__14)
                self.state = 147
                self.sexpr(0)
                self.state = 148
                self.match(IslaLanguageParser.T__16)
                pass

            elif la_ == 13:
                localctx = IslaLanguageParser.SepxrParenContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 150
                self.match(IslaLanguageParser.T__14)
                self.state = 151
                self.sexpr(0)
                self.state = 152
                self.match(IslaLanguageParser.T__16)
                pass

            elif la_ == 14:
                localctx = IslaLanguageParser.SepxrAppContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 154
                self.match(IslaLanguageParser.T__14)
                self.state = 155
                localctx.op = self.sexpr(0)
                self.state = 157 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 156
                    self.sexpr(0)
                    self.state = 159 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << IslaLanguageParser.T__6) | (1 << IslaLanguageParser.T__14) | (1 << IslaLanguageParser.T__17) | (1 << IslaLanguageParser.T__18) | (1 << IslaLanguageParser.T__19) | (1 << IslaLanguageParser.T__20) | (1 << IslaLanguageParser.T__21) | (1 << IslaLanguageParser.T__22) | (1 << IslaLanguageParser.XPATHEXPR) | (1 << IslaLanguageParser.NONTERMINAL_PH) | (1 << IslaLanguageParser.STRING_PH) | (1 << IslaLanguageParser.DSTRINGS_PH) | (1 << IslaLanguageParser.VAR_TYPE) | (1 << IslaLanguageParser.STRING) | (1 << IslaLanguageParser.ID) | (1 << IslaLanguageParser.INT) | (1 << IslaLanguageParser.DIV) | (1 << IslaLanguageParser.MUL) | (1 << IslaLanguageParser.PLUS) | (1 << IslaLanguageParser.MINUS) | (1 << IslaLanguageParser.GEQ) | (1 << IslaLanguageParser.LEQ) | (1 << IslaLanguageParser.GT) | (1 << IslaLanguageParser.LT))) != 0)):
                        break

                self.state = 161
                self.match(IslaLanguageParser.T__16)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 179
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,20,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 177
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
                    if la_ == 1:
                        localctx = IslaLanguageParser.SexprInfixReStrContext(self, IslaLanguageParser.SexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_sexpr)
                        self.state = 165
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
                        self.state = 166
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==IslaLanguageParser.T__21 or _la==IslaLanguageParser.T__22):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 167
                        self.sexpr(7)
                        pass

                    elif la_ == 2:
                        localctx = IslaLanguageParser.SexprInfixPlusMinusContext(self, IslaLanguageParser.SexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_sexpr)
                        self.state = 168
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 169
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==IslaLanguageParser.PLUS or _la==IslaLanguageParser.MINUS):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 170
                        self.sexpr(6)
                        pass

                    elif la_ == 3:
                        localctx = IslaLanguageParser.SexprInfixMulDivContext(self, IslaLanguageParser.SexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_sexpr)
                        self.state = 171
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 172
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==IslaLanguageParser.DIV or _la==IslaLanguageParser.MUL):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 173
                        self.sexpr(5)
                        pass

                    elif la_ == 4:
                        localctx = IslaLanguageParser.SexprInfixEqContext(self, IslaLanguageParser.SexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_sexpr)
                        self.state = 174
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 175
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << IslaLanguageParser.T__6) | (1 << IslaLanguageParser.GEQ) | (1 << IslaLanguageParser.LEQ) | (1 << IslaLanguageParser.GT) | (1 << IslaLanguageParser.LT))) != 0)):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 176
                        self.sexpr(4)
                        pass

             
                self.state = 181
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,20,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class PredicateArgContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(IslaLanguageParser.ID, 0)

        def VAR_TYPE(self):
            return self.getToken(IslaLanguageParser.VAR_TYPE, 0)

        def INT(self):
            return self.getToken(IslaLanguageParser.INT, 0)

        def STRING(self):
            return self.getToken(IslaLanguageParser.STRING, 0)

        def XPATHEXPR(self):
            return self.getToken(IslaLanguageParser.XPATHEXPR, 0)

        def NONTERMINAL_PH(self):
            return self.getToken(IslaLanguageParser.NONTERMINAL_PH, 0)

        def STRING_PH(self):
            return self.getToken(IslaLanguageParser.STRING_PH, 0)

        def DSTRINGS_PH(self):
            return self.getToken(IslaLanguageParser.DSTRINGS_PH, 0)

        def getRuleIndex(self):
            return IslaLanguageParser.RULE_predicateArg

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicateArg" ):
                listener.enterPredicateArg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicateArg" ):
                listener.exitPredicateArg(self)




    def predicateArg(self):

        localctx = IslaLanguageParser.PredicateArgContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_predicateArg)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 182
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << IslaLanguageParser.XPATHEXPR) | (1 << IslaLanguageParser.NONTERMINAL_PH) | (1 << IslaLanguageParser.STRING_PH) | (1 << IslaLanguageParser.DSTRINGS_PH) | (1 << IslaLanguageParser.VAR_TYPE) | (1 << IslaLanguageParser.STRING) | (1 << IslaLanguageParser.ID) | (1 << IslaLanguageParser.INT))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[2] = self.formula_sempred
        self._predicates[3] = self.sexpr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def formula_sempred(self, localctx:FormulaContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 8)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 7)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 6)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 4)
         

    def sexpr_sempred(self, localctx:SexprContext, predIndex:int):
            if predIndex == 5:
                return self.precpred(self._ctx, 6)
         

            if predIndex == 6:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 7:
                return self.precpred(self._ctx, 4)
         

            if predIndex == 8:
                return self.precpred(self._ctx, 3)
         




